/*	Java Phoenix Emulator v0.30
Official Home-page: http://www.dcc.ufmg.br/~muriloq/phoenix

  Emulator by Murilo Saraiva de Queiroz (muriloq@dcc.ufmg.br) based on 
  Phoenix Emulator by Richard Davies (R.Davies@dcs.hull.ac.uk) and MAME 
  project, by Nicola Salmoria (MC6489@mclink.it) and others.
  
	The emulator structure, and many solutions are based in Jasper, the 
	Java Spectrum Emulator, by Adam Davidson & Andrew Pollard.
	Used with permission.
	
	  The machine architecture information is from Ralph Kimmlingen
	  (ub2f@rz.uni-karlsruhe.de).
	  
*/

/** Phoenix Hardware Specification
Resolution 26x8 = 208 columns x 32x8 = 256 lines
Phoenix memory map

  0000-3fff 16Kb Program ROM
  4000-43ff 1Kb Video RAM Charset A (4340-43ff variables)
  4400-47ff 1Kb Work RAM
  4800-4bff 1Kb Video RAM Charset B (4840-4bff variables)
  4c00-4fff 1Kb Work RAM
  5000-53ff 1Kb Video Control write-only (mirrored)
  5400-47ff 1Kb Work RAM
  5800-5bff 1Kb Video Scroll Register (mirrored)
  5c00-5fff 1Kb Work RAM
  6000-63ff 1Kb Sound Control A (mirrored)
  6400-67ff 1Kb Work RAM
  6800-6bff 1Kb Sound Control B (mirrored)
  6c00-6fff 1Kb Work RAM
  7000-73ff 1Kb 8bit Game Control read-only (mirrored)
  7400-77ff 1Kb Work RAM
  7800-7bff 1Kb 8bit Dip Switch read-only (mirrored)
  7c00-7fff 1Kb Work RAM
  
	memory mapped ports:
	
	  read-only:
	  7000-73ff IN
	  7800-7bff DSW
	  
		* IN (all bits are inverted)
		* bit 7 : barrier
		* bit 6 : Left
		* bit 5 : Right
		* bit 4 : Fire
		* bit 3 : -
		* bit 2 : Start 2
		* bit 1 : Start 1
		* bit 0 : Coin
		
		  * DSW
		  * bit 7 : VBlank
		  * bit 6 : free play (pleiads only)
		  * bit 5 : attract sound 0 = off 1 = on (pleiads only?)
		  * bit 4 : coins per play	0 = 1 coin	1 = 2 coins
		  * bit 3 :\ bonus
		  * bit 2 :/ 00 = 3000	01 = 4000  10 = 5000  11 = 6000
		  * bit 1 :\ number of lives
		  * bit 0 :/ 00 = 3  01 = 4  10 = 5  11 = 6
		  
			Pallete
			//0 	]
			//1 	] bit 5-7 of video ram value (divides 256 chars in 8 color sections)
			//2 	]
			//3 ] 2 bit pixelcolor
			//4 ] (either from CHAR-A or CHAR-B, depends on Bit5)
			//5 0= CHAR-A, 1= CHAR-B
			//6 palette flag (see video control reg.)
			//7 always 0
*/

import java.awt.*;
import java.awt.image.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;
import java.net.*;


public class Phoenix extends i8080 {
	public Graphics 	parentGraphics = null;	// applet 
	public Graphics 	canvasGraphics = null;	// emulator window 
	public Graphics 	bufferGraphics = null;	// 
	public Image		bufferImage	   = null;
	
	public Applet		applet = null;
	public AudioClip laserSFX = null;
	public AudioClip explosionSFX = null;
	public AudioClip blowSFX = null;
	public AudioClip shieldSFX = null;
	public AudioClip hitSFX = null;
	public int savedHiScore=0;
	
	public int nPixelsWide = 208;
	public int nPixelsHigh = 256;
	public int pixelScale = 1;
	
	public Container	parent		= null; // PhoenixApp usually (where border is drawn)
	public Canvas		canvas		= null; // Main screen
	
	private byte[] chr = new byte [0x2000]; // CHARSET roms
	
	private int 	DipSwitchSYNC	= 0;		// Atualiza tela
	private int 	ScrollReg		= 0;
	private int 	OldScrollReg	= 0;
	private int 	PalleteReg = 0;
	private boolean BackgroundRefresh = true;
	private int[]   BackgroundBytes = new int [1024];	// Dirty bytes array
	private boolean ScrollRefresh = true;
	private int[]	GameControl = new int [8];
	private int 	interruptCounter = 0;
	private int		frameSkip = 1;
	private int		sleepHack = 5;
	private boolean resetAtNextInterrupt = false;
	private boolean pauseAtNextInterrupt = false;
	private boolean refreshNextInterrupt = true;
	
	public	Thread	pausedThread = null;
	public	long	timeOfLastInterrupt = 0;
	private long	timeOfLastSample = 0;
	public	long	timeOfLastFrameInterrupt = 0;
	private long	timeOfLastFrameSample = 0;
	private float	frameRate;
	// Alpha - Red - Green - Blue
	Color BLACK		= new Color (0x00000000);
	Color WHITE		= new Color (0xffdbdbdb);
	Color RED		= new Color (0xffff0000);
	Color GREEN		= new Color (0xff00ff00);
	Color BLUE		= new Color (0xff2424db);
	Color CYAN		= new Color (0xff00ffdb);
	Color YELLOW	= new Color (0xffffff00);
	Color PINK		= new Color (0xffffb6db);
	Color ORANGE	= new Color (0xffffb649);
	Color LTPURPLE	= new Color (0xffff24b6);
	Color DKORANGE	= new Color (0xffffb600);
	Color DKPURPLE	= new Color (0xffb624ff);
	Color DKCYAN	= new Color (0xff00dbdb);
	Color DKYELLOW	= new Color (0xffdbdb00);
	Color BLUISH	= new Color (0xff9595ff);
	Color PURPLE	= new Color (0xffff00ff);
	
	// pallete x charset x character = color
	/* 4 colors per pixel * 8 groups of characters * 2 charsets * 2 pallettes */
	Color ColorTable[]={
		/* charset A pallette A */	
		BLACK,BLACK,CYAN,CYAN,			// Background, Unused, Letters, asterisks    
			BLACK,YELLOW,RED,WHITE, 		// Background, Ship middle, Numbers/Ship, Ship edge 
			BLACK,YELLOW,RED,WHITE, 		// Background, Ship middle, Ship, Ship edge/bullets 	   
			BLACK,PINK,PURPLE,YELLOW,		// Background, Bird eyes, Bird middle, Bird Wings 	   
			BLACK,PINK,PURPLE,YELLOW,		// Background, Bird eyes, Bird middle, Bird Wings 	   
			BLACK,PINK,PURPLE,YELLOW,		// Background, Bird eyes, Bird middle, Bird Wings 	   
			BLACK,WHITE,PURPLE,YELLOW,		// Background, Explosions 	   
			BLACK,PURPLE,GREEN,WHITE,		// Background, Barrier 
			/* charset A pallette B */
			BLACK,BLUE,CYAN,CYAN,			// Background, Unused, Letters, asterisks 
			BLACK,YELLOW,RED,WHITE, 		// Background, Ship middle, Numbers/Ship, Ship edge 
			BLACK,YELLOW,RED,WHITE, 		// Background, Ship middle, Ship, Ship edge/bullets 	   
			BLACK,YELLOW,GREEN,PURPLE,		// Background, Bird eyes, Bird middle, Bird Wings 	   
			BLACK,YELLOW,GREEN,PURPLE,		// Background, Bird eyes, Bird middle, Bird Wings 	   
			BLACK,YELLOW,GREEN,PURPLE,		// Background, Bird eyes, Bird middle, Bird Wings 	   
			BLACK,WHITE,RED,PURPLE, 		// Background, Explosions 	   
			BLACK,PURPLE,GREEN,WHITE,		// Background, Barrier 
			/* charset B pallette A */
			BLACK,RED,BLUE,WHITE,			// Background, Starfield 
			BLACK,PURPLE,BLUISH,DKORANGE,	// Background, Planets 	  
			BLACK,DKPURPLE,GREEN,DKORANGE,	// Background, Mothership: turrets, u-body, l-body 
			BLACK,BLUISH,DKPURPLE,LTPURPLE, // Background, Motheralien: face, body, feet
			BLACK,PURPLE,BLUISH,GREEN,		// Background, Eagles: face, body, shell   
			BLACK,PURPLE,BLUISH,GREEN,		// Background, Eagles: face, body, feet	  
			BLACK,PURPLE,BLUISH,GREEN,		// Background, Eagles: face, body, feet	  
			BLACK,PURPLE,BLUISH,GREEN,		// Background, Eagles: face, body, feet
			/* charset B pallette B */
			BLACK,RED,BLUE,WHITE,			// Background, Starfield 
			BLACK,PURPLE,BLUISH,DKORANGE,	// Background, Planets 	  
			BLACK,DKPURPLE,GREEN,DKORANGE,	// Background, Mothership: turrets, upper body, lower body 
			BLACK,BLUISH,DKPURPLE,LTPURPLE, // Background, Motheralien: face, body, feet 
			BLACK,BLUISH,LTPURPLE,GREEN,	// Background, Eagles: face, body, shell  
			BLACK,BLUISH,LTPURPLE,GREEN,	// Background, Eagles: face, body, feet 	  
			BLACK,BLUISH,LTPURPLE,GREEN,	// Background, Eagles: face, body, feet 	  
			BLACK,BLUISH,LTPURPLE,GREEN,	// Background, Eagles: face, body, feet 
	};
	public Phoenix ( Container _parent ) throws Exception {
		// Phoenix runs at 0.74 Mhz (?)
		super(0.74);
		
		parent = _parent;
		
		
		parent.add( canvas = new Canvas() );
		canvas.resize( nPixelsWide*pixelScale, nPixelsHigh*pixelScale );
		canvas.show();
		canvasGraphics = canvas.getGraphics();
		
		bufferImage = parent.createImage( nPixelsWide*pixelScale*4, nPixelsHigh*pixelScale );
		bufferGraphics = bufferImage.getGraphics();
		
		for (int i=0;i<8;i++) GameControl[i]=1;
		canvasGraphics.setColor (BLACK);
		bufferGraphics.setColor (BLACK);
		canvasGraphics.fillRect(0,0,nPixelsWide*pixelScale,nPixelsHigh*pixelScale);
		bufferGraphics.fillRect(0,0,nPixelsWide*pixelScale*4,nPixelsHigh*pixelScale);
		canvasGraphics.setColor (Color.yellow);
		bufferGraphics.setColor (Color.yellow);
		
	}
	
	
	/** Byte access */
	public void pokeb( int addr, int newByte ) {
		
		addr &= 0xffff;
		
		if (addr >=  0x5800 && addr <= 0x5bff) {
			ScrollReg = newByte;		// E' preciso repetir em pokew ?
			if (ScrollReg!=OldScrollReg) {
				OldScrollReg=ScrollReg;
				ScrollRefresh = true;
				
			}
			
		}
		if ((addr >= 0x4800)&&(addr <= 0x4bff)) {
			BackgroundRefresh = true;  
		}
		
		if (addr >= 0x5000 && addr <= 0x53ff ) {
			PalleteReg = newByte & 0x01;
		}
		
		if (addr >= 0x6000 && addr <= 0x63ff ) {
			if (peekb(addr)!=newByte) {
				mem[addr] = newByte;
				if (newByte==143) {
					explosionSFX.play ();
					
				}
				if ((newByte>101)&&(newByte<107)) {
					laserSFX.play ();
				}
				
				
				if (newByte==80) {
					blowSFX.play ();
					
				}
				
				// System.out.println ("Sound Control A:"+newByte);
				
			}
			
		}
		
		if (addr >= 0x6800 && addr <= 0x6bff ) {
			if (peekb(addr)!=newByte) {
				mem[ addr ] = newByte;
				if (newByte==12) {
					shieldSFX.play ();
				}
				
				if (newByte==2) {
					hitSFX.play ();
				}
				 // System.out.println ("Sound Control B:"+newByte);
				
			}
			
		}
		
		// Hi Score Saving - Thanks MAME ! :)
		if (addr == 0x438c) {
			if (newByte == 0x0f) {
				mem[addr]=newByte;
				int hiScore = getScore(0x4388);
				if (hiScore > savedHiScore) hisave();
				if (hiScore < savedHiScore) hiload();
				
			}
			
			
		}

		if ( addr >= 0x4000 ) { 	// 0x0000 - 0x3fff Program ROM 
			mem [addr]=newByte;	
		}
		
		return;
		
	}
	
	/** Word access */
	public void pokew( int addr, int word ) {
		addr &= 0xffff;
		int _mem[] = mem;

		if ( addr >= 0x4000 ) {
			_mem[ addr ] = word & 0xff;
			if ( ++addr != 65536 ) {
				_mem[ addr ] = word >> 8;
			}
		
		}
		
		return;
	}
	
	public int peekb( int addr ) {
		addr &= 0xffff;
		// decode the addresses (addr)
		if(addr >= 0x7800 && addr <=0x7bff) //are we reading dip switch memory?
		{
			if(DipSwitchSYNC == 1)			//if SYNC bit of switch is 1
			{
				DipSwitchSYNC = 0;			//set it to 0
				return 128; 				//return value where bit 7 is 1
			}
			else
				return 0;					//return value where bit 7 is 0
		}
		if (addr >= 0x7000 && addr <= 0x73ff) {
			int c=0;
			for (int i=0;i<8;i++) c|=GameControl[i]<<i;
			return c;
		}
		
		else  return mem[addr];
		
	}
	
	public int peekw( int addr ) {
		addr &= 0xffff;
		// decode the addresses (addr)
		if(addr >= 0x7800 && addr <=0x7bff)	//are we reading dip switch memory?
		{
			if(DipSwitchSYNC == 1)			//if SYNC bit of switch is 1
			{
				DipSwitchSYNC = 0;			//set it to 0
				return 128; 				//return value where bit 7 is 1
			}
			else
				return 0;					//return value where bit 7 is 0
		}
		
		int 	   t = peekb( addr );
		addr++;
		return t | (peekb( addr ) << 8);
	}
	
	
	public void initSFX(Applet a) { 
		// Sound Effects loading
		applet = a;
		URL baseURL = applet.getDocumentBase();
		laserSFX = applet.getAudioClip(baseURL,"laser.au");
		laserSFX.play();
		laserSFX.stop();
		
		explosionSFX = applet.getAudioClip(baseURL,"explo.au");
		explosionSFX.play();
		explosionSFX.stop();
		
		blowSFX = applet.getAudioClip(baseURL,"blow.au");
		blowSFX.play();
		blowSFX.stop();
		
		shieldSFX = applet.getAudioClip(baseURL,"shield.au");
		shieldSFX.play();
		shieldSFX.stop();

		hitSFX = applet.getAudioClip(baseURL,"hit.au");
		hitSFX.play();
		hitSFX.stop();
		
	}
	
	// The Hi Score is BCD (Binary Coded Decimal).
	// We convert this to integer here.
	public int getScore(int Addr) {
		int score=0;
		score += (int) (peekb (Addr+3)/16) * 10			+ (peekb (Addr+3) % 16);
		score += (int) (peekb (Addr+2)/16) * 1000		+ (peekb (Addr+2) % 16)*100;
		score += (int) (peekb (Addr+1)/16) * 100000		+ (peekb (Addr+1) % 16)*10000;
		score += (int) (peekb (Addr)  /16) * 10000000	+ (peekb (Addr) % 16)*1000000;
		return score;
	}
	
	public void hisave () {
	// Hi score saving. Again, thanks MAME project... :)
		int OneScore=getScore(0x4380);
		int TwoScore=getScore(0x4384);
		int HiScore=getScore(0x4388);
		int HiAddress = 0x4388;
		if (OneScore > HiScore) HiAddress=0x4380;
		if (TwoScore > HiScore) HiAddress=0x4384;
		
		
		try {
			// URL baseURL = applet.getDocumentBase();
			OutputStream os;
			/*
			if (baseURL != null) {
			URL scoreURL = new URL (baseURL, "hiscore.sav");
			URLConnection connection = new URLConnection (scoreURL);
			os = connection.getOutputStream();				
			}
			else {
			*/
			File scoreFile = new File ("hiscore.sav");
			os = new FileOutputStream (scoreFile);
			//}
			for (int i=0;i<4;i++) {
				os.write ((byte) peekb(HiAddress+i));
			}
				os.flush();
				os.close();
			} catch (Exception e) {
				System.out.println ("Error saving high score");
			}
			savedHiScore = getScore (HiAddress);
			System.out.println ("High Score: "+savedHiScore+" saved.");
		}
		
		public void hiload () {
			int HiAddress = 0x4388;
			try {
				// URL baseURL = applet.getDocumentBase();
				InputStream is;
				/*
				if (baseURL != null) {
				URL scoreURL = new URL (baseURL, "hiscore.sav");
				URLConnection connection = new URLConnection (scoreURL);
				os = connection.getOutputStream();				
				}
				else {
				*/
				File scoreFile = new File ("hiscore.sav");
				is = new FileInputStream (scoreFile);
				//}
				for (int i=0;i<4;i++) mem [HiAddress+i]=is.read ();
				is.close();
			} catch (Exception e) {
				System.out.println ("Error loading high score");
			}
			// Force hi score atualizing
			pokeb(0x41e1, (peekb(0x4389) / 16)+0x20);
			pokeb(0x41c1, (peekb(0x4389) & 0xf)+0x20);
			pokeb(0x41a1, (peekb(0x438a) / 16)+0x20);
			pokeb(0x4181, (peekb(0x438a) & 0xf)+0x20);
			pokeb(0x4161, (peekb(0x438b) / 16)+0x20);
			pokeb(0x4141, (peekb(0x438b) & 0xf)+0x20);

			savedHiScore = getScore(HiAddress);
			System.out.println ("High Score: "+savedHiScore+" loaded.");
		}
		
		/** Carga das roms */
		public void loadRom (URL baseURL) {
			byte buffer[] = new byte [0x4000];
			int readbytes=0;
			System.out.print ("Reading program roms... ");
			try {
				InputStream is;
				if (baseURL != null) {
					URL romURL = new URL (baseURL, "program.rom");
					is = romURL.openStream();				// Abre a rom para leitura
				}
				else {
					File romFile = new File ("program.rom");
					is = new FileInputStream (romFile);
				}
				BufferedInputStream bis = new BufferedInputStream (is, 0x4000);
				int n=0x4000;
				int toRead = 0x4000;
				while ( toRead > 0 ) {
					int nRead = bis.read( buffer, n-toRead, toRead );
					toRead -= nRead;
					readbytes += nRead;
				}
				System.out.println (readbytes+" bytes");
				for (int i=0;i<=0x3fff;i++) {mem[i]=(buffer[i]+256)&0xff;}
			}
			catch (Exception e) {
				System.out.println ("Error loading Phoenix program rom");
			}
		}
		
		/** Carga dos bancos de caracteres*/
		public void loadChr (URL baseURL) {
			byte buffer[] = new byte [0x2000];
			int readbytes=0;
			System.out.print ("Reading character set roms... ");
			
			try {
				InputStream is;
				if (baseURL != null) {
					URL romURL = new URL (baseURL, "graphics.rom");
					is = romURL.openStream();				// Abre a rom para leitura
				}
				else {
					File romFile = new File ("graphics.rom");
					is = new FileInputStream (romFile);
				}
				BufferedInputStream bis = new BufferedInputStream (is, 0x2000);
				int n=0x2000;
				int toRead = 0x2000; 
				while ( toRead > 0 ) {
					int nRead = bis.read( buffer, n-toRead, toRead );
					toRead -= nRead;
					readbytes += nRead;
				}
				System.out.println (readbytes+" bytes");				
				for (int i=0;i<=0x1fff;i++) {chr[i]=buffer[i];}
			}
			catch (Exception e) {
				System.out.println ("Error loading Phoenix graphics rom");
			}
		}
		
		
		
		
		public final int interrupt() {
			if ( pauseAtNextInterrupt ) {
				pausedThread = Thread.currentThread();
				while ( pauseAtNextInterrupt ) {
					try { Thread.sleep( 500 ); } catch ( Exception ignored ) {}
				}	
			}
			pausedThread = null;
			
			if ( resetAtNextInterrupt ) {
				resetAtNextInterrupt = false;
				reset();
			}
			
			interruptCounter++;
			
			DipSwitchSYNC = 1;
			
			if (interruptCounter % frameSkip == 0)  screenRefresh ();
			/*
			// Update speed indicator every 2 seconds of 'phoenix time'
			if ( (interruptCounter % 100 ) == 0 ) {
			timeOfLastFrameSample = System.currentTimeMillis();
			frameRate =(float) ( 1/  ( ((float)timeOfLastFrameSample-(float)timeOfLastFrameInterrupt)/100)) * 1000   ;
			System.out.println ((int)frameRate);
			timeOfLastFrameInterrupt = timeOfLastFrameSample;
			//refreshSpeed();
			}
			*/
			timeOfLastInterrupt = System.currentTimeMillis();
			
			// Trying to slow to 100%, browsers resolution on the system
			// time is not accurate enough to check every interrurpt. So
			// we check every 4 interrupts.
			if ( (interruptCounter % 4) == 0 ) {
				long durOfLastInterrupt = timeOfLastInterrupt - timeOfLastSample;
				timeOfLastSample = timeOfLastInterrupt;
				if ( /*!runAtFullSpeed && */(durOfLastInterrupt < 40) ) {
					try { Thread.sleep( 60 - durOfLastInterrupt ); }
					catch ( Exception ignored ) {}
				}
			}
			
			// This was put in to handle Netscape 2 which was prone to
			// locking up if one thread never gave up its timeslice.
			/* if ( sleepHack > 0 ) {
			try { Thread.sleep( sleepHack ); }
			catch ( Exception ignored ) {}
			}
			*/
			
			return super.interrupt();
		}
		
		public void screenRefresh () {
			int x,y,c0,c1,pos0, pos1;
			int ox0, oy0, ox1, oy1;
			bufferGraphics.setColor(BLACK);
			bufferGraphics.fillRect(0,0,nPixelsWide,nPixelsHigh);
			if (BackgroundRefresh) {
				
				for(y = 0; y < 32; y++) {			// rows of characters
					
					pos1 = 0x4800+32*(26-1)+y;		// position into video memory
					for (x = 0; x < 26; x++) {		// columns of characters
						
						
						c1 = mem[pos1]; 			// get character at current pos
						
						ox1 = PalleteReg*nPixelsWide+nPixelsWide+(c1%26)*8; 
						oy1 = ((int)(c1/26))*8;
						
						
						if ((c1!=BackgroundBytes[pos1-0x4800])) {
							
							bufferGraphics.copyArea(ox1,oy1,8,8,x*8-ox1+624,y*8-oy1); // background
							BackgroundBytes[pos1-0x4800]= c1;
						}
						
						pos1 -= 32; 				 //move to next character in VRAM
					}								 //end for
					
				}									 //end for
				BackgroundRefresh = false;
			};
			
			bufferGraphics.copyArea(624,0,nPixelsWide,ScrollReg,-624,nPixelsHigh-ScrollReg);
			bufferGraphics.copyArea(624,ScrollReg,nPixelsWide,nPixelsHigh-ScrollReg,-624,-ScrollReg);
			ScrollRefresh=false;
			
			for(y = 0; y < 32; y++) {				//rows of characters
				pos0 = 0x4000+32*(26-1)+y;			//position into video memory
				
				for (x = 0; x < 26; x++) {			//columns of characters
					
					c0 = mem[pos0]; 				//get character at current pos
					
					ox0 = PalleteReg*nPixelsWide+nPixelsWide+(c0%26)*8;
					oy0 = ((int)(c0/26))*8+160;
					
					if ((c0!=0)&&(c0!=45)&&(c0!=46))
						bufferGraphics.copyArea(ox0,oy0,8,8,x*8-ox0,y*8-oy0); // upper plane
					pos0 -= 32; 				 //move to next character in VRAM
					
				}								//end for
				
			}									//end for
			
			canvasGraphics.drawImage(bufferImage,0,0,null);
		}
		
		public void decodeChars () {
		/* Due the lack of explicit pointers, and the inefficiency of the use of a 
		great number of Image objects, I've decided to apply an non-conventional
		approach. All game graphics is kept in a invisible region of the buffer 
		image. When one is necessary, I've simply copy the image from a region to
		another. It's not elegant, nor efficient: unfortunately, it's the only 
		choice I Know... :)
			*/
			
			
			
			byte bite;							// Byte lido da rom
			byte[] Bin= new byte[8];			// Byte em binrio
			byte [][] block = new byte[8][8];
			int pixelColorIndex;
			System.out.print ("Decoding character sets to off-screen buffer...");
			
			for (int it=0;it<8;it++) {
				for (int jt=0;jt<8;jt++) {
					block[it][jt]=0;
				}
			}
			for (int s=0;s<2;s++) { 						// Charset
				
				
				for (int c=0;c<256;c++) {					// Character
					
					for (int it=0;it<8;it++) {
						for (int jt=0;jt<8;jt++) {
							block[it][jt]=0;
						}
					}		
					for (int plane=0; plane<2;plane++) {	// Plane
						
						for (int line=0; line<8; line++) {	// Linha
							
							bite = (byte) chr[s*4096+c*8+plane*256*8+line];
							Bin[0]=(byte) ((bite & 1)>>0);
							Bin[1]=(byte) ((bite & 2)>>1);
							Bin[2]=(byte) ((bite & 4)>>2);
							Bin[3]=(byte) ((bite & 8)>>3);
							Bin[4]=(byte) ((bite & 16)>>4);
							Bin[5]=(byte) ((bite & 32)>>5);
							Bin[6]=(byte) ((bite & 64)>>6);
							Bin[7]=(byte) ((bite & 128)>>7);
							
							for (int col=0;col<8;col++) { 	// Coluna
								block[line][col]+= (1+(1-plane))*Bin[col];
								if (plane==1) {
									
									pixelColorIndex = (( (c & 0xff) >> 5 )&0xff)*4;			// bits 5-7 of video ram value
									pixelColorIndex += block[line][col];					// pixel color
									pixelColorIndex += (1-s) * 64;							// Charset
									
									if ((block[line][col]>0)) {
										canvasGraphics.setColor (ColorTable[pixelColorIndex]);
									} else{
										canvasGraphics.setColor (BLACK);
									}
									bufferGraphics.setColor (ColorTable[pixelColorIndex]);
									//}
									
									canvasGraphics.fillRect(7-line+(c%26)*8,col+((int)c/26)*8+s*160,1,1);
									bufferGraphics.fillRect(nPixelsWide+7-line+(c%26)*8,col+((int)c/26)*8+s*160,1,1);
									
									if ((block[line][col]>0)) {
										canvasGraphics.setColor (ColorTable[pixelColorIndex+32]);
									} else{
										canvasGraphics.setColor (BLACK);
									}
									bufferGraphics.fillRect(nPixelsWide+nPixelsWide+7-line+(c%26)*8,col+((int)c/26)*8+s*160,1,1);
									
								} // if
								
							} // for col
							
						} // for line
					} // for plane		
				} // for c
				
			} // for s
			
			System.out.println ("Ok."); 	
			for (int i=0;i<1024;i++) BackgroundBytes[i] = mem[i+0x4800];
			timeOfLastInterrupt = System.currentTimeMillis();	// Inicia calculo do frame rate
			
		}
		/** Process events from UI */
		public boolean handleEvent( Event e ) {
			switch ( e.id ) {
			case Event.MOUSE_DOWN:
				canvas.requestFocus();
				return true;
			case Event.KEY_ACTION:
			case Event.KEY_PRESS:
				return doKey( 1, e.key);
			case Event.KEY_ACTION_RELEASE:
			case Event.KEY_RELEASE:
				return doKey( 0, e.key);
			case Event.GOT_FOCUS:
			case Event.LOST_FOCUS:
				for (int i=0;i<8;i++) GameControl[i]=1;;	 // Reseta os controles
				return true;
			}
			
			return false;
		}
		
		public final boolean doKey( int down, int ascii) {
			
			switch ( ascii ) {
			case '3':	 GameControl[0]=1-down;    break; // Coin
			case '1':	 GameControl[1]=1-down;    break; // Start 1
			case '2':	 GameControl[2]=1-down;    break; // Start 2
			case 32 :	 GameControl[4]=1-down;    break; // Fire 
				// RIGHT = 1007 LEFT = 1006 DOWN = 1005
			case 1007:	 GameControl[5]=1-down;    break; // Right
			case 1006:	 GameControl[6]=1-down;    break; // Left
			case 1005:	 GameControl[7]=1-down;    break; // Barrier
				
			case 'b':	 GameControl[7]=1-down;    break; // Barrier
				
			case '[':	 frameSkip-=down;
				if (frameSkip < 1) frameSkip = 1;	// Decrease frame rate
				System.out.println ("Frame Skip:"+frameSkip);break;
			case ']':	 frameSkip+=down;		    // Increase frame rate
				System.out.println ("Frame Skip:"+frameSkip);
				break;
			}
			
			
			return true;
			
		}
		
		
}











